@@ -1,4 +1,3 @@
-#include <zlib.h>
 #include "gcab-priv.h"
 
 static voidpf
@@ -136,6 +135,8 @@ hexdump (guchar *p, gsize s)
     g_debug ("%15s: %s", #field, p->field)
 #define PN(p, field, size) \
     g_debug ("%15s:", #field), hexdump (p->field, size)
+#define PND(p, field, size) \
+    g_debug ("%15s:", #field), hexdump (field, size)
 
 #define W1(val) \
     g_data_output_stream_put_byte (out, val, cancellable, error)
@@ -185,6 +186,7 @@ cheader_read (cheader_t *ch, GDataInputStream *in,
         goto end;
     }
 
+    memset (ch, 0, sizeof (cheader_t));
     R4 (ch->res1);
     R4 (ch->size);
     R4 (ch->res2);
@@ -400,21 +402,40 @@ cdata_write (cdata_t *cd, GDataOutputStream *out, int type,
     return TRUE;
 }
 
+G_GNUC_INTERNAL void
+cdata_finish (cdata_t *cd, GError **error)
+{
+    z_stream *z = &cd->z;
+    int zret;
+
+    if (!z->opaque)
+        return;
+
+    zret = inflateEnd (z);
+    z->opaque = NULL;
+
+    if (zret != Z_OK)
+        g_set_error (error, GCAB_ERROR, GCAB_ERROR_FAILED,
+                     "zlib failed: %s", zError (zret));
+}
+
 G_GNUC_INTERNAL gboolean
 cdata_read (cdata_t *cd, u1 res_data, GCabCompression compression,
             GDataInputStream *in, GCancellable *cancellable, GError **error)
 
 {
     gboolean success = FALSE;
+    int zret = Z_OK;
+    gchar *data = compression == GCAB_COMPRESSION_NONE ? cd->out : cd->data;
 
     R4 (cd->checksum);
     R2 (cd->ncbytes);
     R2 (cd->nubytes);
     cd->reserved = g_malloc (res_data);
     RN (cd->reserved, res_data);
-    RN (cd->data, cd->ncbytes);
+    RN (data, cd->ncbytes);
 
-    CHECKSUM datacsum = compute_checksum(cd->data, cd->ncbytes, 0);
+    CHECKSUM datacsum = compute_checksum(data, cd->ncbytes, 0);
     g_return_val_if_fail (cd->checksum == compute_checksum ((guint8*)&cd->ncbytes, 4, datacsum), FALSE);
 
     if (g_getenv ("GCAB_DEBUG")) {
@@ -424,11 +445,62 @@ cdata_read (cdata_t *cd, u1 res_data, GCabCompression compression,
         P2 (cd, nubytes);
         if (res_data)
             PN (cd, reserved, res_data);
-        PN (cd, data, 64);
+        PND (cd, data, 64);
+    }
+
+    if (compression == GCAB_COMPRESSION_MSZIP) {
+        z_stream *z = &cd->z;
+
+        if (cd->data[0] != 'C' || cd->data[1] != 'K')
+            goto end;
+
+        z->avail_in = cd->ncbytes - 2;
+        z->next_in = cd->data + 2;
+        z->avail_out = cd->nubytes;
+        z->next_out = cd->out;
+
+        if (!z->opaque) {
+            z->zalloc = zalloc;
+            z->zfree = zfree;
+            z->opaque = cd;
+
+            zret = inflateInit2 (z, -15);
+            if (zret != Z_OK)
+                goto end;
+        }
+
+        while (1) {
+            zret = inflate (z, Z_BLOCK);
+            if (zret != Z_OK)
+                break;
+        }
+
+        if (zret != Z_STREAM_END)
+            goto end;
+
+        g_warn_if_fail (z->avail_in == 0);
+        g_warn_if_fail (z->avail_out == 0);
+        if (z->avail_in != 0 || z->avail_out != 0)
+            goto end;
+
+        zret = inflateReset (z);
+        if (zret != Z_OK)
+            goto end;
+
+        zret = inflateSetDictionary (z, cd->out, cd->nubytes);
+        if (zret != Z_OK)
+            goto end;
     }
 
     success = TRUE;
 
 end:
+    if (zret != Z_OK)
+        g_set_error (error, GCAB_ERROR, GCAB_ERROR_FAILED,
+                     "zlib failed: %s", zError (zret));
+    if (!*error && !success)
+        g_set_error (error, GCAB_ERROR, GCAB_ERROR_FAILED,
+                     "Invalid cabint chunk");
+
     return success;
 }
