@@ -1,5 +1,25 @@
 #include "gcab-priv.h"
 
+/**
+ * SECTION:gcab-folder
+ * @title: GCabFolder
+ * @short_description: A Cabinet folder
+ * @see_also: #GCabFolder
+ * @stability: Stable
+ * @include: libgcab.h
+ *
+ * A GCabFolder is a handle to a folder within the Cabinet archive. A
+ * Cabinet folder <emphasis>is not</emphasis> like a directory. It is
+ * a sub-container grouping GCabFiles together, sharing some common
+ * settings like the compression method.
+ *
+ * You can retrieve the files withing a folder with
+ * gcab_folder_get_files().
+ *
+ * In order to add a file to a folder for creation, use
+ * gcab_folder_add_file().
+ */
+
 struct _GCabFolderClass
 {
     GObjectClass parent_class;
@@ -30,6 +50,8 @@ gcab_folder_finalize (GObject *object)
     g_hash_table_unref (self->hash);
     if (self->reserved)
         g_byte_array_unref (self->reserved);
+    if (self->stream)
+        g_object_unref (self->stream);
 
     G_OBJECT_CLASS (gcab_folder_parent_class)->finalize (object);
 }
@@ -222,7 +244,9 @@ gcab_folder_add_file (GCabFolder *self, GCabFile *file,
  * gcab_folder_get_nfiles:
  * @cabfolder: a #GCabFolder
  *
- * Returns: Number of files in this @folder.
+ * Get the number of files in this @folder.
+ *
+ * Returns: a #guint
  **/
 guint
 gcab_folder_get_nfiles (GCabFolder *self)
@@ -234,7 +258,12 @@ gcab_folder_get_nfiles (GCabFolder *self)
 
 /**
  * gcab_folder_new:
- * @compression: compression used by the folder
+ * @compression: compression to used in this folder
+ *
+ * Creates a new empty Cabinet folder. Use gcab_folder_add_file() to
+ * add files to an archive.
+ *
+ * A Cabinet folder is not a file path, it is a container for files.
  *
  * Returns: a new #GCabFolder
  **/
@@ -247,17 +276,23 @@ gcab_folder_new (GCabCompression compression)
 }
 
 G_GNUC_INTERNAL GCabFolder *
-gcab_folder_new_with_cfolder (const cfolder_t *folder)
+gcab_folder_new_with_cfolder (const cfolder_t *folder, GInputStream *stream)
 {
-    return g_object_new (GCAB_TYPE_FOLDER,
-                         "compression", folder->typecomp,
-                         NULL);
+    GCabFolder *self = g_object_new (GCAB_TYPE_FOLDER,
+                                     "compression", folder->typecomp,
+                                     NULL);
+    self->stream = g_object_ref (stream);
+    self->cfolder = *folder;
+
+    return self;
 }
 
 /**
  * gcab_folder_get_files:
  * @cabfolder: a #GCabFolder
  *
+ * Get the list of #GCabFile files contained in the @cabfolder.
+ *
  * Returns: (element-type GCabFile) (transfer full): list of files
  **/
 GSList *
@@ -267,3 +302,111 @@ gcab_folder_get_files (GCabFolder *self)
 
     return g_slist_reverse (g_slist_copy (self->files));
 }
+
+static gint
+sort_by_offset (GCabFile *a, GCabFile *b)
+{
+    g_return_val_if_fail (a != NULL, 0);
+    g_return_val_if_fail (b != NULL, 0);
+
+    return (gint64)a->cfile.uoffset - (gint64)b->cfile.uoffset;
+}
+
+G_GNUC_INTERNAL gboolean
+gcab_folder_extract (GCabFolder *self,
+                     GFile *path,
+                     u1 res_data,
+                     GCabFileCallback file_callback,
+                     GFileProgressCallback progress_callback,
+                     gpointer callback_data,
+                     GCancellable *cancellable,
+                     GError **error)
+{
+    GError *my_error = NULL;
+    gboolean success = FALSE;
+    GDataInputStream *data = NULL;
+    GOutputStream *out = NULL;
+    GSList *f, *files = NULL;
+    cdata_t cdata = { 0, };
+
+    data = g_data_input_stream_new (self->stream);
+    g_data_input_stream_set_byte_order (data, G_DATA_STREAM_BYTE_ORDER_LITTLE_ENDIAN);
+    g_filter_input_stream_set_close_base_stream (G_FILTER_INPUT_STREAM (data), FALSE);
+
+    if (!g_seekable_seek (G_SEEKABLE (data), self->cfolder.offsetdata, G_SEEK_SET, cancellable, error))
+        goto end;
+
+    files = g_slist_sort (g_slist_copy (self->files), (GCompareFunc)sort_by_offset);
+
+    u4 nubytes = 0;
+    for (f = files; f != NULL; f = f->next) {
+        GCabFile *file = f->data;
+
+        if (file_callback && !file_callback (file, callback_data))
+            continue;
+
+        gchar *fname = g_strdup (gcab_file_get_extract_name (file));
+        int i = 0, len = strlen (fname);
+        for (i = 0; i < len; i++)
+            if (fname[i] == '\\')
+                fname[i] = '/';
+
+        GFileOutputStream *out;
+        GFile *gfile = g_file_resolve_relative_path (path, fname);
+        GFile *parent = g_file_get_parent (gfile);
+        g_free (fname);
+
+        if (!g_file_make_directory_with_parents (parent, cancellable, &my_error)) {
+            if (g_error_matches (my_error, G_IO_ERROR, G_IO_ERROR_EXISTS))
+                g_clear_error (&my_error);
+            else {
+                g_object_unref (gfile);
+                g_object_unref (parent);
+                g_propagate_error (error, my_error);
+                goto end;
+            }
+        }
+        g_object_unref (parent);
+
+        out = g_file_replace (gfile, NULL, FALSE, 0, cancellable, error);
+        g_object_unref (gfile);
+        if (!out)
+            goto end;
+
+        u4 usize = file->cfile.usize;
+        u4 uoffset = file->cfile.uoffset;
+        do {
+            if ((nubytes + cdata.nubytes) <= uoffset) {
+                nubytes += cdata.nubytes;
+                if (!cdata_read (&cdata, res_data, self->compression,
+                                 data, cancellable, error))
+                    goto end;
+                continue;
+            } else {
+                gsize offset = file->cfile.uoffset > nubytes ?
+                    file->cfile.uoffset - nubytes : 0;
+                const void *p = &cdata.out[offset];
+                gsize count = MIN (usize, cdata.nubytes - offset);
+                if (!g_output_stream_write_all (G_OUTPUT_STREAM (out), p, count,
+                                                NULL, cancellable, error))
+                    goto end;
+                usize -= count;
+                uoffset += count;
+            }
+        } while (usize > 0);
+    }
+
+    success = TRUE;
+
+end:
+    if (files)
+        g_slist_free (files);
+    if (data)
+        g_object_unref (data);
+    if (out)
+        g_object_unref (out);
+
+    cdata_finish (&cdata, NULL);
+
+    return success;
+}
